import 'package:viqcore_flutter/src/http_client/utils/dio_helper.dart';

import '../app.dart';

/// A callback that returns a Dio response, presumably from a Dio method
/// it has called which performs an HTTP request, such as `dio.get()`,
/// `dio.post()`, etc.
typedef HttpLibraryMethod<T> = Future<Response<T>> Function();

typedef ApiResponseTransformer<T> = T Function(dynamic data);

class ApiService {
  ApiService({required this.dio});

  final Dio dio;

  /// HTTP GET request.
  Future<ApiResponse<T>> get<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onReceiveProgress,
  }) {
    return request(
      'GET',
      path,
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );
  }

  /// HTTP POST request.
  Future<ApiResponse<T>> post<T>(
    String path, {
    dynamic data,
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
  }) {
    return request(
      'POST',
      path,
      data: data,
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
    );
  }

  Future<ApiResponse<T>> request<T>(String method, String path,
      {dynamic data,
      Map<String, dynamic>? queryParameters,
      Options? options,
      CancelToken? cancelToken,
      ProgressCallback? onSendProgress,
      ProgressCallback? onReceiveProgress,
      ApiResponseTransformer? transformer}) {
    return _mapApiResponse<T, dynamic>(
        () => dio.request(
              path,
              data: data,
              queryParameters: queryParameters,
              options: (options ?? Options())..method = method,
              cancelToken: cancelToken,
              onSendProgress: onSendProgress,
              onReceiveProgress: onReceiveProgress,
            ),
        transformer: transformer);
  }

  Future<ApiResponse<T>> gql<T>(String method, dynamic query,
      {String? path,
      Map<String, dynamic>? variables,
      Options? options,
      CancelToken? cancelToken,
      ProgressCallback? onSendProgress,
      ProgressCallback? onReceiveProgress,
      ApiResponseTransformer? transformer}) {
    String bodyQuery;

    if (query is GraphQLQuery) {
      bodyQuery = query.build();
    } else {
      bodyQuery = '$query';
    }

    final body = '$method { $bodyQuery }';

    options ??= Options();
    options.extra ??= {};
    options.extra!.addAll({'GQL': true});

    return _mapApiResponse<T, dynamic>(
        () => dio.post(
              path ?? '/graphql',
              data: {'query': body, 'variables': variables},
              options: options,
              cancelToken: cancelToken,
              onSendProgress: onSendProgress,
              onReceiveProgress: onReceiveProgress,
            ),
        transformer: transformer);
  }

  Future<ApiResponse<T>> _mapApiResponse<T, R>(HttpLibraryMethod<R> method,
      {ApiResponseTransformer? transformer}) async {
    try {
      var response = await method();
      var data = response.data;
      return ApiResponse.fromResponse(
        response,
        error: response.gqlErrors ? response.gqlErrorMessage : null,
        data: transformer?.call(data) ?? data,
      );
    } on Exception catch (exception) {
      return ApiResponse.fromException(exception);
    }
  }
}
